import numpy as np
import random
import jax
import jax.numpy as jnp
import jax_unirep.utils as unirep
import jax_unirep.layers as unirep_layer
from functools import partial

ALPHABET_Unirep = [
    "-",
    "M",
    "R",
    "H",
    "K",
    "D",
    "E",
    "S",
    "T",
    "N",
    "Q",
    "C",
    "U",
    "G",
    "P",
    "A",
    "V",
    "I",
    "F",
    "Y",
    "W",
    "L",
    "O",
    "X",
    "start",
    "stop",
]
ALPHABET = [
    "A",
    "R",
    "N",
    "D",
    "C",
    "Q",
    "E",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
]
A2U = np.zeros((len(ALPHABET), len(ALPHABET_Unirep)))
for i, s in enumerate(ALPHABET):
    A2U[i, ALPHABET_Unirep.index(s)] = 1
start_vec = np.zeros((1, 26))
start_vec[0, ALPHABET_Unirep.index("start")] = 1


def decode_useq(s):
    indices = jnp.argmax(s, axis=1)
    return [ALPHABET_Unirep[int(i)] for i in indices][1:]


def decode_seq(s):
    indices = jnp.argmax(s, axis=1)
    return [ALPHABET[i] for i in indices]


def encode_seq(s):  # s is a list
    e = np.zeros((len(s), len(ALPHABET)))
    e[np.arange(len(s)), [ALPHABET.index(si) for si in s]] = 1
    return e


def seq2useq(e):
    return jnp.vstack((start_vec, e @ A2U))


def differentiable_jax_unirep(ohc_seq):
    emb_params = unirep.load_embedding()
    seq_embedding = jnp.stack([jnp.matmul(ohc_seq, emb_params)], axis=0)
    _, mLSTM_apply_fun = unirep_layer.mLSTM(1900)
    weight_params = unirep.load_params()[1]
    h_final, _, outputs = jax.vmap(partial(mLSTM_apply_fun, weight_params))(
        seq_embedding
    )
    h_avg = jnp.mean(outputs, axis=1)
    return h_avg


def resample(key, y, output_shape, nclasses=10):
    """
    Resample the given y-vector to have a uniform classes,
    where the classes are chosen via histogramming y.
    """
    if type(output_shape) is int:
        output_shape = (output_shape,)
    if len(y.shape) == 1:
        # regression
        _, bins = np.histogram(y, bins=nclasses)
        classes = np.digitize(y, bins)
    elif len(y.shape) == 2:
        # classification
        classes = np.argmax(y, axis=1)
        nclasses = y.shape[1]
    else:
        raise ValueError("y must rank 1 or 2")
    uc = np.unique(classes)
    nclasses = uc.shape[0]
    if nclasses == 1:
        return jax.random.choice(key, np.arange(y.shape[0]), shape=output_shape)
    idx = [np.where(classes == uc[i])[0] for i in range(nclasses)]
    c = jax.random.choice(key, np.arange(nclasses), shape=output_shape)
    keys = jax.random.split(key, nclasses)
    f = np.vectorize(lambda i: jax.random.choice(keys[i], idx[i]))
    return f(c)


def transform_var(s):
    # heuristic to make MLP output better behaved.
    return jax.nn.softplus(s) + 1e-6



from wazy.mlp import NaiveBlock
import haiku as hk
from .mlp import EnsembleBlock
from .seq import SeqpropBlock
from .utils import differentiable_jax_unirep, seq2useq, transform_var, ALPHABET
import jax.numpy as jnp
import jax


def model_reduce(out):
    mu = jnp.mean(out[..., 0], axis=0)
    var = jnp.mean(transform_var(out[..., 1]) + out[..., 0] ** 2, axis=0) - mu**2
    epi_var = jnp.std(out[..., 0], axis=0) ** 2
    return mu, var, epi_var


def tree_transpose(list_of_trees):
    """Convert a list of trees of identical structure into a single tree of arrays."""
    return jax.tree_util.tree_map(lambda *xs: jnp.array(xs), *list_of_trees)


class EnsembleModel:
    def __init__(self, config):
        def full_model_forward(x, training=True):
            e = EnsembleBlock(config)
            return e(x, training)

        def model_forward(x, training=False):
            x_dim = tuple([1 for i in range(x.ndim)])
            s = jnp.tile(x, (config.model_number, *x_dim))
            mean, var, epi_var = model_reduce(full_model_forward(s, training=training))
            return mean, var, epi_var

        def model_uncertainty_eval(x, training=False):
            x_dim = tuple([1 for i in range(x.ndim)])
            s = jnp.tile(x, (config.model_number, *x_dim))
            out = full_model_forward(s, training=training)
            epistemic = jnp.std(out[..., 0], axis=0)
            aleatoric = jnp.mean(jax.nn.softplus(out[..., 1]) + 1e-6, axis=0)
            return epistemic, aleatoric  # for each x[i]

        def seq_forward(x, training=True):  # params is trained mlp params
            s = SeqpropBlock()(x)
            if config.pretrained:
                us = seq2useq(s)
                u = differentiable_jax_unirep(us)
            else:
                u = s.flatten()
            mean, var, epi_var = model_forward(u, training=training)
            # We only use epistemic uncertainty, since this is used in BO
            return mean, epi_var

        # transform functions
        self.infer_t = hk.transform(model_forward)
        self.train_t = hk.transform(full_model_forward)
        self.seq_t = hk.transform(seq_forward)
        self.var_t = hk.transform(model_uncertainty_eval)

    def seq_apply(self, params, key, x, training=False):
        """Apply the seqprop model by merging the sequence and trainable parameters"""
        mp = hk.data_structures.merge(params, x[1])
        return self.seq_t.apply(mp, key, x[0], training=training)

    def seq_partition(self, params):
        """Extract the seqprop parameters from the parameters"""
        return hk.data_structures.partition(lambda m, *_: "seqprop" in m, params)[0]

    def random_seqs(self, key, batch_size, params, length, start_seq=None):
        """Generate a batch of tuples of sequences and seqprop r,b parameters

        The params should contain the seqprop block. All others will be ignored.
        """
        if start_seq is None:
            start_seq = jnp.zeros((length, len(ALPHABET)))
        sp = self.seq_partition(params)
        return (
            start_seq
            + jax.random.normal(key, shape=(batch_size, length, len(ALPHABET))),
            tree_transpose(
                [jax.tree_util.tree_map(lambda x: x, sp) for _ in range(batch_size)]
            ),
        )


def build_naive_e2e():
    def naive_model_forward(x):
        e = NaiveBlock()
        return e(x)

    def seq_forward(x):
        s = SeqpropBlock()(x)
        us = seq2useq(s)
        u = differentiable_jax_unirep(us)
        out = naive_model_forward(u)
        return out

    naive_model_forward_t = hk.transform(naive_model_forward)
    naive_seq_t = hk.transform(seq_forward)
    return naive_model_forward_t, naive_seq_t



import os
from glob import glob
from setuptools import setup

exec(open("wazy/version.py").read())

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name="wazy",
    version=__version__,
    description="Pretrained Bayesian Optimization of Sequences",
    author="Ziyue Yang, Andrew White",
    author_email="andrew.white@rochester.edu",
    url="http://thewhitelab.org/Software",
    license="MIT",
    packages=["wazy"],
    install_requires=[
        "jax",
        "dm-haiku",
        "optax",
        "numpy",
        "jax-unirep",
    ],
    test_suite="tests",
    zip_safe=True,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)
